/* gpecomp.c : an ELF file packer for the GP2X handheld

   GPEComp v0.3
   Copyright (C) 2005-2006 Mario Rodriguez (rlyeh)
   All Rights Reserved.

   UCL data compression library.
   Copyright (C) 1996-2004 Markus Franz Xaver Johannes Oberhumer
   All Rights Reserved.

   The UCL library is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of
   the License, or (at your option) any later version.

   The UCL library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the UCL library; see the file COPYING.
   If not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   Markus F.X.J. Oberhumer
   <markus@oberhumer.com>
   http://www.oberhumer.com/opensource/ucl/

   rlyeh
   http://www.retrodev.info
 */


/*************************************************************************
// NOTE: this is an example program, so do not use to backup your data.
//
// This program lacks things like sophisticated file handling but is
// pretty complete regarding compression - it should provide a good
// starting point for adaption for you applications.
**************************************************************************/

#include <ucl/ucl.h>
#if defined(UCL_USE_ASM)
#  include <ucl/ucl_asm.h>
#endif
#if 1 && defined(UCL_USE_ASM)
   /* use assembler versions of the decompressors */
#  define ucl_nrv2b_decompress_8        ucl_nrv2b_decompress_asm_8
#  define ucl_nrv2b_decompress_safe_8   ucl_nrv2b_decompress_asm_safe_8
#  define ucl_nrv2d_decompress_8        ucl_nrv2d_decompress_asm_8
#  define ucl_nrv2d_decompress_safe_8   ucl_nrv2d_decompress_asm_safe_8
#  define ucl_nrv2e_decompress_8        ucl_nrv2e_decompress_asm_8
#  define ucl_nrv2e_decompress_safe_8   ucl_nrv2e_decompress_asm_safe_8
#endif

/* benchmark setup - do not introduce floating point dependencies for DOS16 */
#if !defined(WITH_TIMER) && !defined(UCL_OS_DOS16)
#  define WITH_TIMER 1
#endif

/* portability layer */
#define WANT_UCL_MALLOC 1
#define WANT_UCL_FREAD 1
#if defined(WITH_TIMER)
#define WANT_UCL_UCLOCK 1
#endif
#define WANT_UCL_WILDARGV 1
#include "examples/portab.h"


static const char *progname = NULL;

static unsigned long total_in = 0;
static unsigned long total_out = 0;
static int opt_debug = 0;

/* use option `-F' for faster operation - then we don't compute or verify
 * a checksum and always use the fast decompressor */
static ucl_bool opt_fast = 0;

/* magic file header for compressed files */
static const unsigned char magic[8] =
    { 0x00, 0xe9, 0x55, 0x43, 0x4c, 0xff, 0x01, 0x1a };

/* benchmark */
#if defined(WITH_TIMER)
static ucl_uclock_handle_t uc;
static double benchmark_bytes;
static double benchmark_secs;
#endif


/*************************************************************************
// file IO
**************************************************************************/

ucl_uint xread(FILE *f, ucl_voidp buf, ucl_uint len, ucl_bool allow_eof)
{
    ucl_uint l;

    l = (ucl_uint) ucl_fread(f,buf,len);
    if (l > len)
    {
        fprintf(stderr,"\nsomething's wrong with your C library !!!\n");
        exit(1);
    }
    if (l != len && !allow_eof)
    {
        fprintf(stderr,"\nread error - premature end of file\n");
        exit(1);
    }
    total_in += l;
    return l;
}

ucl_uint xwrite(FILE *f, const ucl_voidp buf, ucl_uint len)
{
    ucl_uint l;

    if (f != NULL)
    {
        l = (ucl_uint) ucl_fwrite(f,buf,len);
        if (l != len)
        {
            fprintf(stderr,"\nwrite error [%ld %ld]  (disk full ?)\n",
                   (long)len, (long)l);
            exit(1);
        }
    }
    total_out += len;
    return len;
}


int xgetc(FILE *f)
{
    unsigned char c;
    xread(f,(ucl_voidp) &c,1,0);
    return c;
}

void xputc(FILE *f, int c)
{
    unsigned char cc = (unsigned char) c;
    xwrite(f,(const ucl_voidp) &cc,1);
}

/* read and write portable 32-bit integers */

ucl_uint32 xread32(FILE *f)
{
    unsigned char b[4];
    ucl_uint32 v;

    xread(f,b,4,0);
    v  = (ucl_uint32) b[3] <<  0;
    v |= (ucl_uint32) b[2] <<  8;
    v |= (ucl_uint32) b[1] << 16;
    v |= (ucl_uint32) b[0] << 24;
    return v;
}

void xwrite32(FILE *f, ucl_uint32 v)
{
    unsigned char b[4];

    b[3] = (unsigned char) (v >>  0);
    b[2] = (unsigned char) (v >>  8);
    b[1] = (unsigned char) (v >> 16);
    b[0] = (unsigned char) (v >> 24);
    xwrite(f,b,4);
}


/*************************************************************************
// util
**************************************************************************/

static ucl_uint get_overhead(int method, ucl_uint size)
{
    if (method == 0x2b || method == 0x2d || method == 0x2e)
        return size / 8 + 256;
    return 0;
}


static char method_name[64];

static ucl_bool set_method_name(int method, int level)
{
    method_name[0] = 0;
    if (level < 1 || level > 10)
        return 0;
    if (method == 0x2b)
        sprintf(method_name,"NRV2B-99/%d", level);
    else if (method == 0x2d)
        sprintf(method_name,"NRV2D-99/%d", level);
    else if (method == 0x2e)
        sprintf(method_name,"NRV2E-99/%d", level);
    else
        return 0;
    return 1;
}


/*************************************************************************
// compress
**************************************************************************/

int do_compress(FILE *fi, FILE *fo, int method, int level, ucl_uint block_size)
{
    int r = 0;
    ucl_bytep in = NULL;
    ucl_bytep out = NULL;
    ucl_uint in_len;
    ucl_uint out_len;
    ucl_uint32 flags = opt_fast ? 0 : 1;
    ucl_uint32 checksum;
    ucl_uint overhead = 0;
#if defined(WITH_TIMER)
    ucl_uclock_t t_start, t_stop;
#endif

    total_in = total_out = 0;

/*
 * Step 1: write magic header, flags & block size, init checksum
 */
    xwrite(fo,magic,sizeof(magic));
    xwrite32(fo,flags);
    xputc(fo,method);           /* compression method */
    xputc(fo,level);            /* compression level */
    xwrite32(fo,block_size);
    checksum = ucl_adler32(0,NULL,0);

/*
 * Step 2: allocate compression buffers and work-memory
 */
    overhead = get_overhead(method,block_size);
    in = (ucl_bytep) ucl_malloc(block_size);
    out = (ucl_bytep) ucl_malloc(block_size + overhead);
    if (in == NULL || out == NULL)
    {
        printf("%s: out of memory\n", progname);
        r = 1;
        goto err;
    }

/*
 * Step 3: process blocks
 */
    for (;;)
    {
        /* read block */
        in_len = xread(fi,in,block_size,1);
        if (in_len <= 0)
            break;

        /* update checksum */
        if (flags & 1)
            checksum = ucl_adler32(checksum,in,in_len);

#if defined(WITH_TIMER)
        ucl_uclock_read(&uc, &t_start);
#endif

        /* compress block */
        r = UCL_E_ERROR;
        out_len = 0;
        if (method == 0x2b)
            r = ucl_nrv2b_99_compress(in,in_len,out,&out_len,0,level,NULL,NULL);
        else if (method == 0x2d)
            r = ucl_nrv2d_99_compress(in,in_len,out,&out_len,0,level,NULL,NULL);
        else if (method == 0x2e)
            r = ucl_nrv2e_99_compress(in,in_len,out,&out_len,0,level,NULL,NULL);
        if (r == UCL_E_OUT_OF_MEMORY)
        {
            printf("%s: out of memory in compress\n", progname);
            r = 1;
            goto err;
        }
        if (r != UCL_E_OK || out_len > in_len + get_overhead(method,in_len))
        {
            /* this should NEVER happen */
            printf("internal error - compression failed: %d\n", r);
            r = 2;
            goto err;
        }

#if defined(WITH_TIMER)
        ucl_uclock_read(&uc, &t_stop);
        benchmark_bytes += in_len;
        benchmark_secs += ucl_uclock_get_elapsed(&uc, &t_start, &t_stop);
#endif

        /* write uncompressed block size */
        xwrite32(fo,in_len);

        if (out_len < in_len)
        {
            /* write compressed block */
            xwrite32(fo,out_len);
            xwrite(fo,out,out_len);
        }
        else
        {
            /* not compressible - write uncompressed block */
            xwrite32(fo,in_len);
            xwrite(fo,in,in_len);
        }
    }

    /* write EOF marker */
    xwrite32(fo,0);

    /* write checksum */
    if (flags & 1)
        xwrite32(fo,checksum);

    r = 0;
err:
    ucl_free(out);
    ucl_free(in);
    return r;
}


/*************************************************************************
// decompression benchmark
**************************************************************************/

#if defined(WITH_TIMER)

void do_decompress_benchmark(int method, unsigned long benchmark_loops,
                             const ucl_bytep in, ucl_uint in_len,
                             ucl_bytep out, ucl_uint out_len)
{
    ucl_uclock_t t_start, t_stop;

    ucl_uclock_read(&uc, &t_start);
    while (benchmark_loops-- > 0)
    {
        int r = -1;
        ucl_uint new_len = out_len;

#if 1 && defined(UCL_USE_ASM)
        /* use fast assembler versions of the decompressors - see
         * asm/i386/00README.ASM for more info */
        if (method == 0x2b)
            r = ucl_nrv2b_decompress_asm_fast_8(in,in_len,out,&new_len,NULL);
        else if (method == 0x2d)
            r = ucl_nrv2d_decompress_asm_fast_8(in,in_len,out,&new_len,NULL);
        else if (method == 0x2e)
            r = ucl_nrv2e_decompress_asm_fast_8(in,in_len,out,&new_len,NULL);
#else
        if (method == 0x2b)
            r = ucl_nrv2b_decompress_8(in,in_len,out,&new_len,NULL);
        else if (method == 0x2d)
            r = ucl_nrv2d_decompress_8(in,in_len,out,&new_len,NULL);
        else if (method == 0x2e)
            r = ucl_nrv2e_decompress_8(in,in_len,out,&new_len,NULL);
#endif
        if (r != UCL_E_OK || new_len != out_len)
        {
            printf("%s: compressed data violation: error %d (0x%x: %ld/%ld/%ld)\n", progname, r, method, (long) in_len, (long) out_len, (long) new_len);
            printf("%s: unexpected failure in benchmark -- exiting.\n", progname);
            exit(1);
        }
        benchmark_bytes += out_len;
    }
    ucl_uclock_read(&uc, &t_stop);
    benchmark_secs += ucl_uclock_get_elapsed(&uc, &t_start, &t_stop);
}

#endif /* WITH_TIMER */


/*************************************************************************
// decompress / test
//
// We are using in-place decompression here.
**************************************************************************/

int do_decompress(FILE *fi, FILE *fo, unsigned long benchmark_loops)
{
    int r = 0;
    ucl_bytep buf = NULL;
    ucl_uint buf_len;
    unsigned char m [ sizeof(magic) ];
    ucl_uint32 flags;
    int method;
    int level;
    ucl_uint block_size;
    ucl_uint32 checksum;
    ucl_uint overhead = 0;

    total_in = total_out = 0;

/*
 * Step 1: check magic header, read flags & block size, init checksum
 */
    if (xread(fi,m,sizeof(magic),1) != sizeof(magic) ||
        memcmp(m,magic,sizeof(magic)) != 0)
    {
        printf("%s: header error - this file is not compressed by uclpack\n", progname);
        r = 1;
        goto err;
    }
    flags = xread32(fi);
    method = xgetc(fi);
    level = xgetc(fi);
    block_size = xread32(fi);
    overhead = get_overhead(method, block_size);
    if (overhead == 0 || !set_method_name(method, level))
    {
        printf("%s: header error - invalid method %d (level %d)\n",
                progname, method, level);
        r = 2;
        goto err;
    }
    if (block_size < 1024 || block_size > 8*1024*1024L)
    {
        printf("%s: header error - invalid block size %ld\n",
                progname, (long) block_size);
        r = 3;
        goto err;
    }

    #ifndef GP2X_STUB
    printf("%s: block-size is %ld bytes\n", progname, (long)block_size);
    #endif

    checksum = ucl_adler32(0,NULL,0);

/*
 * Step 2: allocate buffer for in-place decompression
 */
    buf_len = block_size + overhead;
    if (benchmark_loops > 0)
    {
        /* cannot use in-place decompression when doing benchmarks */
        buf_len += block_size;
    }
    buf = (ucl_bytep) ucl_malloc(buf_len);
    if (buf == NULL)
    {
        printf("%s: out of memory\n", progname);
        r = 4;
        goto err;
    }

/*
 * Step 3: process blocks
 */
    for (;;)
    {
        ucl_bytep in;
        ucl_bytep out;
        ucl_uint in_len;
        ucl_uint out_len;

        /* read uncompressed size */
        out_len = xread32(fi);

        /* exit if last block (EOF marker) */
        if (out_len == 0)
            break;

        /* read compressed size */
        in_len = xread32(fi);

        /* sanity check of the size values */
        if (in_len > block_size || out_len > block_size ||
            in_len == 0 || in_len > out_len)
        {
            printf("%s: block size error - data corrupted\n", progname);
            r = 5;
            goto err;
        }

        /* place compressed block at the top of the buffer */
        in = buf + buf_len - in_len;
        out = buf;

        /* read compressed block data */
        xread(fi,in,in_len,0);

        if (in_len < out_len)
        {
            /* decompress - use safe decompressor as data might be corrupted */
            ucl_uint new_len = out_len;

            if (method == 0x2b)
            {
                if (opt_fast)
                    r = ucl_nrv2b_decompress_8(in,in_len,out,&new_len,NULL);
                else
                    r = ucl_nrv2b_decompress_safe_8(in,in_len,out,&new_len,NULL);
            }
            else if (method == 0x2d)
            {
                if (opt_fast)
                    r = ucl_nrv2d_decompress_8(in,in_len,out,&new_len,NULL);
                else
                    r = ucl_nrv2d_decompress_safe_8(in,in_len,out,&new_len,NULL);
            }
            else if (method == 0x2e)
            {
                if (opt_fast)
                    r = ucl_nrv2e_decompress_8(in,in_len,out,&new_len,NULL);
                else
                    r = ucl_nrv2e_decompress_safe_8(in,in_len,out,&new_len,NULL);
            }
            if (r != UCL_E_OK || new_len != out_len)
            {
                printf("%s: compressed data violation: error %d (0x%x: %ld/%ld/%ld)\n", progname, r, method, (long) in_len, (long) out_len, (long) new_len);
                r = 6;
                goto err;
            }
            /* write decompressed block */
            xwrite(fo,out,out_len);
            /* update checksum */
            if ((flags & 1) && !opt_fast)
                checksum = ucl_adler32(checksum,out,out_len);
#if defined(WITH_TIMER)
            if (benchmark_loops > 0)
                do_decompress_benchmark(method,benchmark_loops,in,in_len,out,out_len);
#endif
        }
        else
        {
            /* write original (incompressible) block */
            xwrite(fo,in,in_len);
            /* update checksum */
            if ((flags & 1) && !opt_fast)
                checksum = ucl_adler32(checksum,in,in_len);
        }
    }

    /* read and verify checksum */
    if (flags & 1)
    {
        ucl_uint32 c = xread32(fi);
        if (!opt_fast && c != checksum)
        {
            printf("%s: checksum error - data corrupted\n", progname);
            r = 7;
            goto err;
        }
    }

    r = 0;
err:
    ucl_free(buf);
    return r;
}


/*************************************************************************
// misc support functions
**************************************************************************/


/* open input file */
static FILE *xopen_fi(const char *name)
{
    FILE *f;

    f = fopen(name,"rb");
    if (f == NULL)
    {
        printf("%s: cannot open input file %s\n", progname, name);
        exit(1);
    }
#if defined(HAVE_STAT) && defined(S_ISREG)
    {
        struct stat st;
#if defined(HAVE_LSTAT)
        if (lstat(name,&st) != 0 || !S_ISREG(st.st_mode))
#else
        if (stat(name,&st) != 0 || !S_ISREG(st.st_mode))
#endif
        {
            printf("%s: %s is not a regular file\n", progname, name);
            fclose(f);
            exit(1);
        }
    }
#endif
    return f;
}


/* open output file */
static FILE *xopen_fo(const char *name)
{
    FILE *f;

#if 0
    /* this is an example program, so make sure we don't overwrite a file */
    f = fopen(name,"rb");
    if (f != NULL)
    {
        printf("%s: file %s already exists -- not overwritten\n", progname, name);
        fclose(f);
        exit(1);
    }
#endif
    f = fopen(name,"wb");
    if (f == NULL)
    {
        printf("%s: cannot open output file %s\n", progname, name);
        exit(1);
    }
    return f;
}


/*************************************************************************
// main
**************************************************************************/

#ifndef GP2X_STUB
#include "unp_stub.c"
#endif

#define STUBSIZE 43592

int __acc_cdecl_main main(int argc, char *argv[])
{
    int i = 1;
    int r = 0;
    FILE *fi = NULL;
    FILE *fo = NULL;
    const char *in_name = NULL;
    const char *out_name = NULL;
    ucl_bool opt_decompress = 0;
    ucl_bool opt_test = 0;
    int opt_method = 0x2d;
    int opt_level = 7;
    ucl_uint opt_block_size;
    unsigned long opt_decompress_loops = 0;
    const char *s;

    ucl_wildargv(&argc, &argv);
#if defined(WITH_TIMER)
    ucl_uclock_open(&uc);
#endif
    progname = argv[0];
    for (s = progname; *s; s++)
        if ((*s == '/' || *s == '\\') && s[1])
            progname = s + 1;

#ifndef GP2X_STUB

    printf("GPEComp v0.3 (Linux ELF compressor for GP2X) coded by rlyeh, 2005.\nUCL library by Markus F.X.J. Oberhumer\n\n");

    if(argc!=3)
    {
     printf("Usage:\n\n%s infile.gpe outfile.gpe\n",progname);
     return(1);
    }

#endif

/*
 * Step 1: initialize the UCL library
 */
    if (ucl_init() != UCL_E_OK)
    {
        printf("internal error - ucl_init() failed !!!\n");
        printf("(this usually indicates a compiler bug - try recompiling\nwithout optimizations, and enable `-DUCL_DEBUG' for diagnostics)\n");
        exit(1);
    }

/*
 * Step 2: setup default options
 */
    opt_block_size = 256 * 1024L;

#if defined(ACC_MM_AHSHIFT)
    /* reduce memory requirements for ancient 640kB DOS real-mode */
    if (ACC_MM_AHSHIFT != 3)
        opt_block_size = 16 * 1024L;
#endif


/*
 * Step 3: get options
 */

#ifdef GP2X_STUB
    opt_decompress = 1;
    if (opt_decompress)
    {
        char in_name2[255], out_name2[255];

        sprintf(in_name2 ,"%s"             ,progname);
        sprintf(out_name2,"/mnt/tmp/%s_tmp",progname);

        system("mkdir -p /mnt/tmp");
        system("mount -t tmpfs -o rw,size=16M none /mnt/tmp");

        //printf("uncrunching %s %s\n",in_name2, out_name2);

        fi = xopen_fi(in_name2);

        fseek(fi, STUBSIZE, SEEK_SET);

        //printf("current position %lu, ",ftell(fi));
        //printf("value [%02x]\n", fgetc(fi));

        //printf("current position %lu, ",ftell(fi));
        //printf("value [%02x]\n", fgetc(fi));
        //fseek(fi, -2, SEEK_CUR);

        ftell(fi);fgetc(fi);
        ftell(fi);fgetc(fi);
        fseek(fi, -2, SEEK_CUR);

        fo = xopen_fo(out_name2);

        if(fo==NULL) { fclose(fi); return(1); }

        r = do_decompress(fi, fo, 0);

        if (fi) fclose(fi);
        if (fo) fclose(fo);

        if (r == 0)
        {
            chmod(out_name2,777);

                {
                 int argp=1;
                
                 sprintf(in_name2, out_name2);

                 while(argp<argc)
                  sprintf(in_name2, "%s %s",in_name2,argv[argp++]);

                 system(in_name2);
                }

            remove(out_name2);

            system("umount /mnt/tmp");
        }

    }

#else

    {
        //printf("%s: using block-size of %lu bytes\n", progname, (unsigned long)opt_block_size);
        if (!set_method_name(opt_method, opt_level))
        {
            printf("%s: internal error - invalid method %d (level %d)\n",
                   progname, opt_method, opt_level);
            goto quit;
        }
        in_name = argv[1];
        out_name = argv[2];
        fi = xopen_fi(in_name);
        fo = xopen_fo(out_name);
        fwrite(mystub, 1, STUBSIZE, fo);
        r = do_compress(fi,fo,opt_method,opt_level,opt_block_size);
        if (r == 0)
        {
            if(total_in < total_out+STUBSIZE)
            {
            printf("Can't compress this .gpe! It is small enough.\n");
            fclose(fo);
            remove(argv[2]);
            }

            else

            printf("Done! algorithm %s, crunched %lu into %lu bytes\n",
                    method_name, total_in, total_out+STUBSIZE);
#if defined(WITH_TIMER)
            if (opt_debug >= 1 && benchmark_bytes > 0)
            {
                /* speed measured in uncompressed kilobytes per second */
                double kb = benchmark_bytes / (1024.0);
                double kbs = 0.0;
                if (benchmark_secs >= 0.0001) kbs = kb / benchmark_secs;
                printf("\n  [compression speeed] %.3f KB, %.3f secs, %.3f KB/s\n\n", kb, benchmark_secs, kbs);
            }
#endif
        }
    }

#endif



quit:
    if (fi) fclose(fi);
    if (fo) fclose(fo);
#if defined(WITH_TIMER)
    ucl_uclock_close(&uc);
#endif

#ifdef GP2X_STUB
            /* back to gp2x menu */
            chdir("/usr/gp2x");
            execl("gp2xmenu","gp2xmenu",NULL);
#endif

    return r;
}

/*
vi:ts=4:et
*/


